#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#include <VRP.h>

static int port = 4579;
static int i_am_server = 0;
static vrp_incoming_t incoming;
static vrp_outgoing_t outgoing;

static void vrp_test_incoming_handler(vrp_in_buffer_t in_buffer);

static void*vrp_test_background_incoming_poller(void*arg)
{
  vrp_incoming_t i = (vrp_incoming_t)arg;
  while(1)
    {
      vrp_incoming_read(i);
    }
  return NULL;
}

static void*vrp_test_background_outgoing_poller(void*arg)
{
  vrp_outgoing_t o = (vrp_outgoing_t)arg;
  while(1)
    {
      vrp_outgoing_read(o);
    }
  return NULL;
}

int vrp_test_init(int argc, const char**argv)
{
  const char*server = NULL;
  int loss_rate_fact=0;
  int cons_loss_fact=0;
  int cons_loss_max=0;
  
#ifdef PADICO
  server = padico_na_host_getname(padico_na_node_gethost(padico_na_getnodebyrank(0)));
  i_am_server = (padico_na_rank() == 0);
#else /* PADICO */
  if(argc == 1)
    i_am_server = 1;
  else if(argc == 2)
    {
      i_am_server = 0;
      server = argv[1];
    }
  else if (argc == 4)
    {
      i_am_server = 1;
      loss_rate_fact = atoi(argv[1]);
      cons_loss_fact = atoi(argv[2]);
      cons_loss_max  = atoi(argv[3]);
    }
  else
    {
      fprintf(stderr, "usage: %s [<server>]\n", argv[0]);
      fprintf(stderr, "       %s single_loss_rate consecutive_losses_rate max_consecutive_losses\n", argv[0]);
      exit(1);
    }
#endif /* PADICO */

  if(i_am_server)
    {
      pthread_t tid;
      printf("I am server. Creating incoming...\n");
      incoming = vrp_incoming_construct(&port, &vrp_test_incoming_handler);
      printf("incoming ok -- port=%d\n", port);
      if( pthread_create(&tid, NULL, &vrp_test_background_incoming_poller, incoming) )
	{
	  fprintf(stderr, "Ooops can't create a new thread\n");
	  exit(2);
	}
    }
  else
    {      
      printf("I am client. Creating outgoing to %s:%d...\n", server, port);
      outgoing = vrp_outgoing_construct(server, port, 128, 1024,0);
      if(outgoing != NULL)
	printf("outgoing ok...\n");
      else
	{
	  printf("outgoing init failed.\n");
	  exit(1);
	}
      {
	pthread_t tid;
	if( pthread_create(&tid, NULL, &vrp_test_background_outgoing_poller, outgoing) )
	  {
	    fprintf(stderr, "Ooops can't create a new thread\n");
	    exit(2);
	  }
      }
      vrp_outgoing_connect(outgoing);
    }
  return 0;
}

int vrp_test_run(int argc, const char**argv)
{
  const int count = 32;
  const int data_size =  10000000;
  const int max_cons_loss = 10000; /* maximum consecutive loss: 10000 bytes */
  const int max_percent_loss = 10; /* maximum loss per frame: 10% */
  const int min_ref_rate     =  5; /* maximum loss of the same packet : 5 */
  char*data;

  if(i_am_server)
    {
      while(1) sleep(10);
    }

  if(!i_am_server)
    {
      int i;
      struct timeval tv1, tv2;
      double time;

      data = malloc(data_size);
      printf("Starting transfer.\n");
      
      gettimeofday(&tv1, NULL);
      for(i=0; i<count; i++)
	{
	  vrp_buffer_t buffer;
	  printf("Constructing buffer i=%d size=%d\n", i, data_size);
	  buffer = vrp_buffer_construct(data, data_size);
	  printf("Set loss parameters: cons=%d bytes, max=%d%%, min_ref_rate=%d\n",
		 max_cons_loss,
		 max_percent_loss,
		 min_ref_rate);
	  vrp_buffer_set_loss_rate(buffer, max_percent_loss, max_cons_loss);
	  vrp_outgoing_set_min_refresh_rate(outgoing, min_ref_rate);
	  printf("Sending buffer.\n");
	  vrp_outgoing_send_frame(outgoing, buffer); 
	  printf("Destroying buffer.\n");
	  vrp_buffer_destroy(buffer);
	}
      gettimeofday(&tv2, NULL);
      time = (tv2.tv_usec-tv1.tv_usec) / 1000000.0;
      time+= tv2.tv_sec-tv1.tv_sec;
      printf("Total time: %f (%i bytes)\n", time, count * data_size);
      printf("  %f bytes/s\n", data_size*count/time);
      printf("  %f MB/s\n", data_size*count/(1000000*time));
      free(data);
    }
  return 0;
}

void vrp_test_finalize()
{
  if(i_am_server)
    {
      vrp_incoming_close(incoming);
    }
  else
    {
      vrp_outgoing_close(outgoing);
    }
}

static void vrp_test_incoming_handler(vrp_in_buffer_t in_buffer)
{
  int i;
  
  printf("[vrp-test] Frame arrived at %p size=%d\n",
	       in_buffer->data, in_buffer->size);
  printf("[vrp-test]   bytes lost=%d (%g%%)\n",
	       in_buffer->loss_residual, 
	       (100.0*in_buffer->loss_residual)/in_buffer->size);
  printf("[vrp-test]   %d fragments lost, each %d bytes\n",
	        in_buffer->loss_num, in_buffer->loss_granularity);
 
  for(i=0; i<in_buffer->loss_num; i++)
    {
      printf("[vrp-test]   missing fragment #%d %p:%p\n",
		   i, in_buffer->lost[i],
		   in_buffer->lost[i]+in_buffer->loss_granularity);
    }
}

int main(int argc, const char**argv)
{
  vrp_init();
  vrp_test_init(argc, argv);
  vrp_test_run(argc, argv);
  vrp_test_finalize();
  vrp_shutdown();
}
