/** @file
 * @brief Sender-side part of VRP
 * @ingroup VRP
 */

#include "VRP-internals.h"


/* TODO: - determine a reasonable timeout - adaptive? 
 *       - parameterize the window size 
 *       - detect multi-packet frame-header
 */

/* notice: - we assume that VRP_OUTGOING_TIMEOUT < 1,000,000  (1 sec)
 *         - we assume VRP_OUTGOING_WINDOW_SIZE to be an exact power of 2
 */


struct vrp_out_timing_param_s
{
  long int timeout;            /**< timeout (in usec) */
  long int RTT;                /**< smoothed round-trip time (in usec) */
  long int average_deviation;  /**< average deviation of actual RTT against smoothed RTT (in usec) */
};

struct vrp_out_cell_s
{
  unsigned int       acked:   1;
  unsigned int       sent:    1;
  unsigned int       expired: 1;
  struct timeval     expire_date;
  struct timeval     expected;
  struct timeval     start_date;
  struct timeval     ack_time;
  struct vrp_out_timing_param_s timing; /**< timing params snapshot (at packet start) */
};

struct vrp_out_hdr_s
{
  char content[VRP_MAX_PACKET_SIZE];
  int  size; /* in bytes- actual size of the header */
  struct vrp_out_cell_s cell;
};

enum vrp_out_frame_state_e
{
  VRP_OUT_FRAME_NONE = 0,   /**< no frame in this slot */
  VRP_OUT_FRAME_INIT,       /**< sending frame header */
  VRP_OUT_FRAME_SEND,       /**< sending data */
  VRP_OUT_FRAME_REFUSED,    /**< frame refused by incoming */
  VRP_OUT_FRAME_ERROR,      /**< transmission error */
  VRP_OUT_FRAME_CLOSING     /**< all data sent */
};

enum vrp_outgoing_state_e
{
  VRP_OUTGOING_INIT = 0,
  VRP_OUTGOING_OK,
  VRP_OUTGOING_CLOSING,
  VRP_OUTGOING_CLOSED
};

struct vrp_out_frame_s
{
  int frame_num;
  enum vrp_out_frame_state_e state;
  struct timeval deadline;

  int i; /* log => TO REMOVE */

  /* sliding window */
  int ack_expected;
  int next_to_send;
  
  /* buffers */
  int next_to_buffer;
  struct vrp_out_cell_s* *cells;
  int num_cells;
  char send_buf[VRP_MAX_PACKET_SIZE];
  /* data */
  struct vrp_buffer_s*src;
  int max_seq;          /**< maximum packet sequence number for this frame */
  int data_per_packet;  /**< data per packet in bytes */
  /* frame header */
  struct vrp_out_hdr_s hdr;
};
typedef struct vrp_out_frame_s*vrp_out_frame_t;

struct vrp_outgoing_s
{
  enum vrp_outgoing_state_e state;

  int cwnd_size; /* congestion window */
  int fwnd_size; /* flow window */

 /* slow start */
  vrp_mutex_t mutex;
  int ssthresh;
  int ack_cg_num;
  int will_slow_start;
  int will_slow_start_size;
  
  /* system */
  int fd;
#ifdef VRP_PADICO
  padico_io_t io;
#endif
  struct sockaddr_in local_addr;
  struct sockaddr_in outgoing_addr;
  struct hostent*host;
  char * hostname;
  int port; /* in host byte order */

  struct vrp_out_timing_param_s timing;  /**< dynamic timeout adjustment parameters */
#ifdef VEGAS_ENABLED
  long int base_rtt;
  long int avg_rtt_diff;
  long int old_RTT;
#endif
  struct vrp_out_frame_s*current_frame;
  int last_frame_num;

  int max_cons_loss;
  int max_loss_rate;
  int min_ref_rate;

  int n_read;
  char recv_buf[VRP_MAX_PACKET_SIZE];

  int fault_code;
};

/* private functions */

static inline struct vrp_out_cell_s* vrp_out_cell_translate(vrp_out_frame_t frame, int i)
{
  return frame->cells[ (i % frame->num_cells) ];
}

static void
vrp_outgoing_adapt_timeout(vrp_outgoing_t outgoing, 
			   struct vrp_out_cell_s*);

static void
vrp_outgoing_adapt_cwnd_size(vrp_outgoing_t outgoing,
			      int newsize);

static void
vrp_adapt_ssthresh(vrp_outgoing_t outgoing);


static void
vrp_outgoing_slowstart(vrp_outgoing_t outgoing);


static void 
vrp_outgoing_timeout_expired(vrp_outgoing_t outgoing);


static void
vrp_start_timer(struct vrp_out_cell_s*cell, vrp_outgoing_t outgoing);

static int
vrp_check_timeout(struct vrp_out_cell_s*cell);

static void 
vrp_out_frame_alive(vrp_out_frame_t frame);

static int 
vrp_out_frame_isalive(vrp_out_frame_t frame);


static vrp_out_frame_t
vrp_out_frame_construct(vrp_outgoing_t outgoing, 
			vrp_buffer_t buffer);

static void 
vrp_out_frame_destroy(vrp_outgoing_t outgoing, vrp_out_frame_t frame);


static void 
vrp_outgoing_send_packet(vrp_outgoing_t outgoing, 
			 vrp_out_frame_t frame, 
			 int seq);

static void 
vrp_out_frame_send_hdr(vrp_outgoing_t outgoing, 
		       vrp_out_frame_t frame);

static void 
vrp_out_frame_hdr_construct(vrp_outgoing_t outgoing, 
			    vrp_out_frame_t frame);

static void 
vrp_outgoing_process_fragment(vrp_outgoing_t outgoing);

static void 
vrp_outgoing_send_routine(vrp_outgoing_t outgoing, 
			  vrp_out_frame_t frame);

/* *****************************************************************
 * **        VRP outgoing funcs
 * *****************************************************************
 */



/*
 * vrp_outgoing_construct()
 *
 * Construct a vrp_outgoing_t for the given host and port. Look up in the
 * outgoing table to see if one already exists. If it does, bump its reference
 * count and return that one. Otherwise create one, insert into the
 * table with a reference count of 1 and return it.
 */
vrp_outgoing_t
vrp_outgoing_construct(char*hostname, int port, 
		       int max_loss_rate, int max_cons_loss, int min_ref_rate)
{
  vrp_outgoing_t outgoing;
  int rc;
  int len;
  int sock_buffsize;
  int i;
  struct timeval now;

  outgoing = (vrp_outgoing_t)vrp_malloc(sizeof(struct vrp_outgoing_s));
  vrp_show(50, "enter, host = %s, port = %d\n", hostname, port);
  outgoing->state = VRP_OUTGOING_INIT;

  /* *** initialize slow start *** */
  vrp_mutex_init(outgoing->mutex);
  outgoing->ssthresh=VRP_OUTGOING_SLOW_START_INIT_BYTES/VRP_MAX_PACKET_SIZE;
  outgoing->ack_cg_num=0;
  outgoing->will_slow_start=0;
  outgoing->will_slow_start_size=VRP_OUTGOING_SLOW_START_INIT_BYTES/VRP_MAX_PACKET_SIZE;

  outgoing->fwnd_size = VRP_OUTGOING_FWND_SIZE;
  outgoing->cwnd_size = VRP_OUTGOING_CWND_INIT_SIZE;

  /* *** create the socket */
  /* socket */
  rc = vrp_libc_socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
  if(rc < 0) 
    vrp_fatal("socket() [vrp_outgoing_construct()]");
  else
    outgoing->fd = rc;
  vrp_show(60, "socket fd = %i\n", outgoing->fd);
  
  /* bind */
  vrp_show(60, "binding socket %i to local port\n", outgoing->fd);
  len = sizeof(outgoing->local_addr);
  outgoing->local_addr.sin_family = AF_INET;
  outgoing->local_addr.sin_port   = htons(0);
  memset(outgoing->local_addr.sin_zero,0,8);
  outgoing->local_addr.sin_addr.s_addr = INADDR_ANY;
  rc = vrp_libc_bind(outgoing->fd, 
		     (struct sockaddr*) &outgoing->local_addr, 
		     len);
  if(rc < 0)
    vrp_fatal("bind() [vrp_outgoing_construct()]");
  
  rc = vrp_libc_getsockname(outgoing->fd, 
			    (struct sockaddr*) &outgoing->local_addr, 
			    &len);
  if(rc < 0)
    vrp_fatal("getsockname() [vrp_outgoing_construct()]");
  
  vrp_show(50, "looking up host %s:%i\n", hostname, port);
  outgoing->host = gethostbyname(hostname);
  if (outgoing->host==NULL)
    vrp_fatal("Couldn't resolve host \"%s\" [vrp_outgoing_construct()]\n ",hostname);
  outgoing->outgoing_addr.sin_family = AF_INET;
  memcpy(&outgoing->outgoing_addr.sin_addr.s_addr,
	 outgoing->host->h_addr_list[0],
	 outgoing->host->h_length);
  outgoing->outgoing_addr.sin_port = htons(port);
  
  /*
    vrp_show(60, "Setting descriptor options\n");
    rc = vrp_libc_fcntl(outgoing->fd, F_SETFL, O_NONBLOCK);
    if(rc == -1)
    vrp_fatal("fcntl O_NONBLOCK [vrp_outgoing_construct()]");
  */
  
  sock_buffsize = 64 * 1024 ;
  rc = vrp_libc_setsockopt(outgoing->fd, SOL_SOCKET, SO_RCVBUF,
			   (char *) &sock_buffsize, sizeof(sock_buffsize));
  if(rc < 0)
    vrp_fatal("setsockopt SOL_SOCKET [vrp_outgoing_construct()]");
  
  /* *** init internal structures */
  outgoing->current_frame = NULL;
  srand(getpid());
  outgoing->port = port;
  outgoing->hostname = hostname;
  outgoing->max_loss_rate = max_loss_rate * 256 / 100;
  outgoing->max_cons_loss = max_cons_loss;
  outgoing->min_ref_rate = min_ref_rate;
  outgoing->last_frame_num = ((unsigned)rand()) % VRP_FRAME_MAX;
  outgoing->n_read = 0;
  outgoing->timing.timeout = VRP_TIMEOUT_DEFAULT;
  outgoing->timing.RTT = VRP_TIMEOUT_DEFAULT;
  
  gettimeofday(&now, NULL);
#ifdef VEGAS_ENABLED
  outgoing->base_rtt=10000000;
  outgoing->avg_rtt_diff=0;
  outgoing->old_RTT=0;
#endif
  
  outgoing->timing.average_deviation = VRP_TIMEOUT_DEFAULT/4;
  outgoing->fault_code = 0;
  fd_finder.fd_table[outgoing->fd].type = FD_TABLE_TYPE_OUTGOING;
  fd_finder.fd_table[outgoing->fd].value = (void *) outgoing;
#ifdef VRP_PADICO
  outgoing->io = padico_io_register(outgoing->fd,
				    PADICO_IO_EVENT_READ,
				    &vrp_outgoing_read_callback,
				    outgoing);
  padico_io_activate(outgoing->io);
  vrp_outgoing_connect(outgoing);
#endif

  return outgoing;
}


/* *** open the connection: ping */
vrp_outgoing_t 
vrp_outgoing_connect(vrp_outgoing_t outgoing) 
{
  struct vrp_out_hdr_s ping_hdr;
  struct timeval date;
  while(outgoing->state == VRP_OUTGOING_INIT)
    {
      gettimeofday(&date, NULL);
      vrp_hdr_pack_std(ping_hdr.content, &ping_hdr.size);
      vrp_hdr_pack_tag(ping_hdr.content, &ping_hdr.size, VRP_TAG_PING, date.tv_sec, date.tv_usec);
      vrp_hdr_pack_tag(ping_hdr.content, &ping_hdr.size, VRP_TAG_LAST);
      vrp_print("sending ping to IP=%s:%d (size=%d)...\n", 
		inet_ntoa(outgoing->outgoing_addr.sin_addr), 
		ntohs(outgoing->outgoing_addr.sin_port), ping_hdr.size);
      vrp_libc_sendto(outgoing->fd,
		      &ping_hdr.content, 
		      ping_hdr.size, 
		      0,
		      (struct sockaddr*)&outgoing->outgoing_addr, 
		      sizeof(outgoing->outgoing_addr));
      vrp_start_timer(&ping_hdr.cell, outgoing);
      while((!vrp_check_timeout(&ping_hdr.cell)) && (outgoing->state != VRP_OUTGOING_OK))
	{
	  vrp_yield();
	}
    }
#ifdef VEGAS_ENABLED
  outgoing->old_RTT=outgoing->timing.RTT;
  outgoing->base_rtt=outgoing->timing.RTT;
#endif
  vrp_print("Connection established.\n");
  vrp_print("VRP: outgoing statistics\n"
	  "  timeout = %ld usec\n  RTT = %ld usec (latency = %g ms)\n  average_deviation = %ld\n",
	    (long)outgoing->timing.timeout,
	    (long)outgoing->timing.RTT, 
	    (double)(outgoing->timing.RTT/2000.0),
	    (long)outgoing->timing.average_deviation);
  return outgoing;
}

#ifndef VRP_PADICO
int
vrp_outgoing_fd(vrp_outgoing_t outgoing)
{
  return outgoing->fd;
}
#endif 


/*
 * vrp_outgoing_close()
 *
 * Close a outgoing:
 *   1) Remove the outgoing from the fd_table table
 *   2) Close the fd
 *   3) Modify outgoing data structure
 */
void 
vrp_outgoing_close(vrp_outgoing_t outgoing)
{
  int done;
  char ack_buf[VRP_MAX_PACKET_SIZE];
  int ack_size;

  vrp_show(50, "closing outgoing\n");
  outgoing->state = VRP_OUTGOING_CLOSING;

  do
    {
      done = 1;
      if((outgoing->current_frame != NULL) && 
	 (outgoing->current_frame->state != VRP_OUT_FRAME_CLOSING))
	done = 0;
	
    } 
  while(!done);
  vrp_hdr_pack_std(ack_buf, &ack_size);
  /*
  for(i = 0; i < VRP_OUTGOING_NUM_FRAMES; i++)
    {
      if(outgoing->frames[i].state != VRP_OUT_FRAME_NONE)
	{
	  vrp_hdr_pack_tag(ack_buf, &ack_size, VRP_TAG_FRAME_CLOSE, outgoing->frames[i].frame_num);
	  vrp_out_frame_destroy(&outgoing->frames[i]);
	}
    }
  */
  if(outgoing->current_frame != NULL)
    {
      vrp_hdr_pack_tag(ack_buf, &ack_size, VRP_TAG_FRAME_CLOSE, outgoing->current_frame->frame_num);
      vrp_out_frame_destroy(outgoing, outgoing->current_frame);
    }
  
  vrp_hdr_pack_tag(ack_buf, &ack_size, VRP_TAG_LAST);
  vrp_libc_sendto(outgoing->fd,
		  (char*)&ack_buf,
		  ack_size, 
		  0,
		  (struct sockaddr*)&outgoing->outgoing_addr, 
		  sizeof(outgoing->outgoing_addr));
  outgoing->state = VRP_OUTGOING_CLOSED;
#ifdef VRP_PADICO
  padico_io_unregister(outgoing->io);
#endif
  vrp_libc_close(outgoing->fd);
  
  fd_finder.fd_table[outgoing->fd].type = FD_TABLE_TYPE_NONE;
  fd_finder.fd_table[outgoing->fd].value = NULL;
#ifndef VRP_MADELEINE
  outgoing->fd = -1;
  vrp_free(outgoing);
#endif
}
   
void
vrp_outgoing_set_min_refresh_rate(vrp_outgoing_t outgoing, int min_ref_rate)
{
  if (min_ref_rate >= 0) outgoing->min_ref_rate = min_ref_rate;
}

int 
vrp_outgoing_send_frame(vrp_outgoing_t outgoing, 
			vrp_buffer_t buffer)
{
  vrp_out_frame_t frame;
  int rc;

  vrp_show(50, "send frame: entering\n");
  rc = 0;
  vrp_enter();
  if(outgoing->state != VRP_OUTGOING_OK)
    {
      vrp_show(5, "not ready\n");
      rc = VRP_ERROR_NOT_READY;
      vrp_exit();
      goto fn_exit;
    }
  if(buffer->size <= 0)
    {
      vrp_exit();
      goto fn_exit;
    }
  /* lookup an empty frame slot */
  /*  frame = vrp_out_frame_slot_lookup(outgoing); */
  frame = vrp_out_frame_construct(outgoing, buffer);
  if(frame == NULL)
    {
      rc = VRP_ERROR_NOT_READY;
      vrp_exit();
      goto fn_exit;
    }
  vrp_exit();
  /* we have now a new frame structure - build & send frame header
   * -> we can free the global vrp lock.
   */
  vrp_out_frame_hdr_construct(outgoing, frame);
  vrp_out_frame_send_hdr(outgoing, frame);
  vrp_show(60, "frame header sent\n");
  
  while(frame->state != VRP_OUT_FRAME_CLOSING)
    {
      assert((frame->ack_expected <= frame->max_seq) || 
	     (frame->state == VRP_OUT_FRAME_CLOSING));

      switch(frame->state)
	{
	case VRP_OUT_FRAME_NONE:
	  vrp_warning("VRP: bad frame-- cannot send.\n");
	  return VRP_ERROR_BAD_FRAME;
	  break;
	case VRP_OUT_FRAME_SEND:
	  vrp_outgoing_send_routine(outgoing, frame);
	  break;
	case VRP_OUT_FRAME_INIT:
	  if(vrp_check_timeout(&frame->hdr.cell))
	    {
	      //vrp_outgoing_slowstart(outgoing);

	      vrp_warning("Timeout expired for header frame=%d (timeout=%d)\n",
	      	  frame->frame_num, outgoing->timing.timeout);
	      vrp_outgoing_timeout_expired(outgoing);
	      vrp_out_frame_send_hdr(outgoing, frame);
	    }
	  break;
	case VRP_OUT_FRAME_CLOSING:
	  break;
	case VRP_OUT_FRAME_REFUSED:
	  vrp_out_frame_destroy(outgoing, frame);
	  return VRP_ERROR_CONNECTION_REFUSED;
	case VRP_OUT_FRAME_ERROR:
	  vrp_out_frame_destroy(outgoing, frame);
	  return VRP_ERROR_INTERNAL_ERROR;
	}
      if(!vrp_out_frame_isalive(frame))
	{
	  vrp_show(5, "connection timeout\n");
	  vrp_out_frame_destroy(outgoing, frame);
	  return VRP_ERROR_CONNECTION_TIMEOUT;
	}
      //      vrp_yield(); // TODO
    }
  vrp_out_frame_destroy(outgoing, frame);

fn_exit:
  assert(outgoing->current_frame == NULL);
  return rc;
}

/** @brief compute new timing parameters
 * @note assume that the cell is received before the timeout expires
 */
static void 
vrp_outgoing_adapt_timeout(vrp_outgoing_t outgoing, 
			   struct vrp_out_cell_s*cell)
{
  long int deviation;
  long int actual_RTT;
  int i;
  
  deviation = (cell->ack_time.tv_sec - cell->expected.tv_sec) * 1000000 + 
    (cell->ack_time.tv_usec - cell->expected.tv_usec);

  actual_RTT = (cell->ack_time.tv_sec - cell->start_date.tv_sec) * 1000000 + 
    (cell->ack_time.tv_usec - cell->start_date.tv_usec);

  vrp_show(50, "actual_RTT=%i; smoothed params: RTT=%i timeout=%i average_deviation=%d\n",
	     (int)actual_RTT,
	     (int)outgoing->timing.RTT,
	     (int)outgoing->timing.timeout,
	     (int)outgoing->timing.average_deviation);
  /*
  vrp_log(vrp_log_fd, 
	  "%d %d %d %d\n", 
	  (int)actual_RTT,
	  (int)outgoing->timing.RTT, 
	  (int)outgoing->timing.timeout, 
	  (int)outgoing->timing.average_deviation);
  */
  vrp_log(vrp_log_fd, 
	  "%d %d %d %d %d %d %d \n",
	  outgoing->current_frame->i,
	  outgoing->current_frame->frame_num,	  
	  (int)actual_RTT,
	  (int)outgoing->timing.RTT, 
	  (int)outgoing->timing.timeout, 
	  (int)outgoing->timing.average_deviation);
  outgoing->current_frame->i++;

  if((actual_RTT <= cell->timing.timeout) && (actual_RTT > 0))
    {
      /* timeout has not expired: common case */
      outgoing->timing.average_deviation =
	((VRP_TIMEOUT_DEVIATION_SMOOTH_FACTOR - 1) * outgoing->timing.average_deviation + 
	 abs(deviation)) / VRP_TIMEOUT_DEVIATION_SMOOTH_FACTOR;
      outgoing->timing.RTT = 
	((VRP_TIMEOUT_RTT_SMOOTH_FACTOR - 1) * outgoing->timing.RTT + actual_RTT) / VRP_TIMEOUT_RTT_SMOOTH_FACTOR;
      outgoing->timing.timeout = 
	VRP_TIMEOUT_FEEDBACK_FACTOR  * outgoing->timing.RTT + 
	VRP_TIMEOUT_DEVIATION_FACTOR * outgoing->timing.average_deviation;
#ifdef VEGAS_ENABLED
      outgoing->avg_rtt_diff=
	(VRP_OUTGOING_VEGAS_FACTOR - 1 ) * outgoing->avg_rtt_diff +
	VRP_OUTGOING_VEGAS_FACTOR * (actual_RTT-outgoing->old_RTT);
      if (actual_RTT<outgoing->base_rtt)
	outgoing->base_rtt=actual_RTT;
      outgoing->old_RTT=actual_RTT;
#endif
    }
  else if(actual_RTT <= 0)
    {
      /* ACK arrived before the message left... */
      vrp_warning("unexpected ACK -- cannot compute new timeout\n"
		     " (timeout=%i RTT=%i average_deviation=%i deviation=%i)\n", 
		     (int)outgoing->timing.timeout, 
		     (int)outgoing->timing.RTT, 
		     (int)outgoing->timing.average_deviation, 
		     (int)deviation);
    }
  else
    {
      /* timeout had actually expired */
      vrp_warning("trying to compute timing parameters for an expired cell -- detected scheduler race condition.\n");
      vrp_show(10," deviation=%d; actual_RTT=%d; RTT=%d; timeout=%d; average_deviation=%d\n",
		   (int)deviation,
		   (int)actual_RTT,
		   (int)outgoing->timing.RTT, 
		   (int)outgoing->timing.timeout, 
		   (int)outgoing->timing.average_deviation);
    }
}




static void vrp_outgoing_adapt_cwd_size(vrp_outgoing_t outgoing,int newsize)
{
  vrp_out_frame_t frame;
  int oldncells;
  int oldsize;
  struct vrp_out_cell_s* *oldcells;
  struct vrp_out_cell_s* *newcells;
  struct vrp_out_cell_s*cell;
  int i;

  vrp_mutex_lock(outgoing->mutex);
  
  frame = outgoing->current_frame;
  
  if(frame==NULL)
    {
      vrp_show(20,"adapt_windowsize : Oooops trying to resize a NULL window\n");
    }
  else
    {
      oldsize   = outgoing->cwnd_size;
      oldncells = frame->num_cells;
      oldcells  = frame->cells;
      
      if(newsize < 2 )
	{
	  vrp_show(30,"adapt_cwd_size (%d>%d) : cwnd has reached the lower limit (2)\n",oldsize, newsize);
	  newsize=2;
	}
      else if (newsize>frame->max_seq)
	{
	  vrp_show(30,"adapt_cwd_size (%d>%d) : cwnd has reached the upper limit (%d)\n",oldsize, newsize,frame->max_seq);
	  newsize=frame->max_seq;
	}      
      else if(oldsize == newsize)
	{	  
	  vrp_show(40,"adapt_cwd_size (%d>%d) : oldsize=newsize => NOTHING TO DO\n",oldsize, newsize);
	}
      else
	{
	  if(oldsize < newsize)
	    {
	      vrp_show(40,"adapt_windowsize (%d>%d) : I : oldsize<newsize => INCREASE\n",oldsize, newsize);
	      
	      if(newsize>oldncells)
		{
		  vrp_show(50,"adapt_cwd_size (%d>%d) : I : need to rebuild the window)\n",oldsize, newsize);
		  
		  newcells = vrp_malloc(newsize * sizeof(struct vrp_out_cell_s *));
		  
		  vrp_show(60,"adapt_cwd_size (%d>%d) : I : oldtab move > newtab\n",oldsize, newsize);
		  
		  for(i = frame->ack_expected; i < frame->next_to_send; i++)
		    {
		      newcells[ i % newsize ] = vrp_out_cell_translate(frame, i);
		    }
		  
		  vrp_show(60,"adapt_cwd_size (%d>%d) : I : newtab_end init\n",oldsize, newsize);
		  for(;i < frame->ack_expected+newsize; i++)
		    {
		      cell = vrp_malloc(sizeof(struct vrp_out_cell_s));
		      cell->acked = 0;
		      cell->sent = 0;
		      cell->expired = 0;
		      newcells[i % newsize]=cell;
		    }

		  vrp_show(60,"adapt_cwd_size (%d>%d) : free oldtab\n",oldsize, newsize);
		  vrp_free(frame->cells);
		  frame->num_cells=newsize;
		  frame->cells=newcells;
		}	      
	    }
	  if (oldsize > newsize)
	    {
	      vrp_show(40,"adapt_cwd_size (%d>%d) : D : oldsize>newsize => DECREASE\n",oldsize, newsize);
	    }
	  outgoing->cwnd_size=newsize;
	  vrp_show(60,"adapt_cwd_size (%d>%d) : DONE\n",oldsize, newsize);
	}
      
      /* === log the new size of the sliding window === */
      vrp_log(vrp_logw_fd, "%ld %d %d %d\n",
	      frame->next_to_send,frame->frame_num,outgoing->cwnd_size);
      
    }
  vrp_mutex_unlock(outgoing->mutex);
}

static void vrp_adapt_ssthresh(vrp_outgoing_t outgoing)
{
#ifdef VEGAS_ENABLED
  /* outgoing->ssthresh=outgoing->ssthresh*3/4; */
  outgoing->ssthresh=outgoing->cwnd_size*3/4;
  if ( outgoing->ssthresh<40 )
    outgoing->ssthresh=40;  
#else
  /*  outgoing->ssthresh=outgoing->ssthresh/2;*/
   outgoing->ssthresh=outgoing->cwnd_size/2;
#endif
}

static void vrp_outgoing_slowstart(vrp_outgoing_t outgoing)
{
  vrp_show(5,"slowstart\n");
  outgoing->will_slow_start=1;
  
  vrp_adapt_ssthresh(outgoing);
#ifdef VEGAS_ENABLED
  outgoing->will_slow_start_size=outgoing->cwnd_size/2;
  /* vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size/2); */
#else  
  outgoing->will_slow_start_size=VRP_OUTGOING_CWND_INIT_SIZE;
  /* vrp_outgoing_adapt_cwd_size(outgoing,VRP_OUTGOING_CWND_INIT_SIZE);*/
#endif
}

static void 
vrp_outgoing_timeout_expired(vrp_outgoing_t outgoing)
{
  if((outgoing->timing.timeout > 0) && 
     (outgoing->timing.timeout < VRP_TIMEOUT_MAX))
    { 
     if(outgoing->timing.average_deviation < outgoing->timing.RTT/VRP_TIMEOUT_DEVIATION_FACTOR)
	{
	  outgoing->timing.average_deviation *= 1.8;
	}
      else
	{
	  outgoing->timing.average_deviation *= 1.2;
	}
      
    }
  else if(outgoing->timing.timeout <= 0)
    {
      outgoing->timing.RTT = VRP_TIMEOUT_DEFAULT;
    }
  else
    {
      vrp_show(10, " timeout=%d > MAX_TIMEOUT=%d \n",
		 (int)outgoing->timing.timeout, VRP_TIMEOUT_MAX);
      outgoing->timing.average_deviation = VRP_TIMEOUT_MAX / 4;
      outgoing->timing.RTT = VRP_TIMEOUT_MAX;
    }
  outgoing->timing.timeout =
    VRP_TIMEOUT_FEEDBACK_FACTOR  * outgoing->timing.RTT + 
    VRP_TIMEOUT_DEVIATION_FACTOR * outgoing->timing.average_deviation;
  vrp_show(50, " RTT=%i timeout=%i average_deviation=%d\n",
	     (int)outgoing->timing.RTT,
	     (int)outgoing->timing.timeout,
	     (int)outgoing->timing.average_deviation);
}

static void vrp_out_frame_destroy(vrp_outgoing_t outgoing, vrp_out_frame_t frame)
{
  int i;

  vrp_mutex_lock(outgoing->mutex);

  vrp_show(50,"destroying frame\n");
  if(frame != NULL)
    {
      for(i=0;i<frame->num_cells;i++)
	vrp_free(frame->cells[i]);   
      vrp_free(frame->cells);
      vrp_free(frame);
      outgoing->current_frame = NULL;
    }
  else
    {
      vrp_warning("cannot destroy empty frame\n");
    }
  vrp_show(50,"frame destroyed\n");
  
  vrp_mutex_unlock(outgoing->mutex);
}

/*
static struct vrp_out_frame_s* 
vrp_out_frame_lookup(vrp_outgoing_t outgoing, int frame_num)
{
  int i;
  struct vrp_out_frame_s *current = NULL;

  vrp_show(50, "looking up frame %i\n", frame_num);
  for(i = 0; i < VRP_OUTGOING_NUM_FRAMES; i++)
    {
      if(outgoing->frames[i].state != VRP_OUT_FRAME_NONE)
	{
	  current = &outgoing->frames[i];
	  if(frame_num  == current->frame_num )
	    return(current);
	}
    }
  vrp_show(40, "frame not found\n");
  return NULL;
}
*/

static void 
vrp_out_frame_alive(vrp_out_frame_t frame)
{
  gettimeofday(&frame->deadline, NULL);
  frame->deadline.tv_sec += VRP_DEAD_TIMEOUT;
}

static int 
vrp_out_frame_isalive(vrp_out_frame_t frame)
{
  struct timeval time;
  gettimeofday(&time, NULL);
  return(time.tv_sec <= frame->deadline.tv_sec);
}

/*
static struct vrp_out_frame_s* 
vrp_out_frame_slot_lookup(vrp_outgoing_t outgoing)
{
  int i;
  
  vrp_show(60, "looking up slot\n");
  for(i = 0; i < VRP_OUTGOING_NUM_FRAMES; i++)
    if(outgoing->frames[i].state == VRP_OUT_FRAME_NONE)
      return(&outgoing->frames[i]);
  vrp_show(5, "no frame slot available\n");
  return NULL;
}
*/

static vrp_out_frame_t
vrp_out_frame_construct(vrp_outgoing_t outgoing, 
			vrp_buffer_t   buffer)
{
  int i;
  struct vrp_out_cell_s*cell;
  vrp_out_frame_t frame = vrp_malloc(sizeof(struct vrp_out_frame_s));
  
  frame->i=0; /* log => TO REMOVE */

  vrp_show(50, "new frame\n");
  frame->state = VRP_OUT_FRAME_INIT;
  frame->frame_num = outgoing->last_frame_num;
  outgoing->last_frame_num = (outgoing->last_frame_num + 1) % VRP_FRAME_MAX;
  /* build window */
  frame->next_to_send = 0;
  frame->ack_expected = 0;
  frame->next_to_buffer = 0;
  frame->num_cells = VRP_OUTGOING_FWND_SIZE;
  frame->cells = vrp_malloc(frame->num_cells * sizeof(struct vrp_out_cell_s *));
  for(i = 0; i < frame->num_cells; i++)
    {
      cell = vrp_malloc(sizeof(struct vrp_out_cell_s));
      cell->acked = 0;
      cell->sent = 0;
      cell->expired = 0;
      frame->cells[i]=cell;
    }
  /* cut data into packets */
  frame->data_per_packet = VRP_DATA_PER_PACKET;
  frame->max_seq = (buffer->size - 1) / VRP_DATA_PER_PACKET;
  frame->src = buffer;
  vrp_out_frame_alive(frame);
  vrp_show(50, "frame %i : %i packets\n", frame->frame_num, frame->max_seq+1);
  assert(outgoing->current_frame == NULL);
  outgoing->current_frame = frame;
  return frame;
}


static void 
vrp_outgoing_send_packet(vrp_outgoing_t outgoing, 
			 vrp_out_frame_t frame, 
			 int seq)
{
  struct vrp_hdr_s hdr;
  int size;

  vrp_show(50,"seq = %i frame = %i\n", seq, frame->frame_num);
  /* false loss statistics
  vrp_out_cell_translate(frame, seq)->lost = vrp_outgoing_packet_loss(outgoing);  if(vrp_out_cell_translate(frame, seq)->lost)
    {
      hdr.frame = frame->frame_num;
      hdr.seq = 1;
      hdr.is_hdr = 0;
      vrp_hdr_pack(&hdr, frame->send_buf);
      size = ((seq >= frame->max_seq) ? 
	      frame->src->size - frame->data_per_packet * frame->max_seq : 
	      frame->data_per_packet);
      memcpy(frame->send_buf + VRP_HDR_SIZE,
	     frame->src->data + seq * frame->data_per_packet,
	     size);
      size += VRP_HDR_SIZE;
    }
  else
    {
  */

  hdr.frame = frame->frame_num;
  hdr.seq = seq;
  hdr.is_hdr = 0;
  vrp_hdr_pack(&hdr, frame->send_buf);
  size = ((seq >= frame->max_seq) ? 
	  frame->src->size - frame->data_per_packet * frame->max_seq : 
	  frame->data_per_packet);
  memcpy(frame->send_buf + VRP_HDR_SIZE,
	 frame->src->data + seq * frame->data_per_packet,
	 size);
  size += VRP_HDR_SIZE;

      /*    }*/
  vrp_libc_sendto(outgoing->fd,
		  (char*)&frame->send_buf,
		  size, 
		  0,
		  (struct sockaddr*)&outgoing->outgoing_addr, 
		  sizeof(outgoing->outgoing_addr));
  
  if(vrp_out_cell_translate(frame, seq)->sent)
    {
      vrp_out_cell_translate(frame, seq)->expired = 1;
    }
  else
    {
      vrp_out_cell_translate(frame, seq)->sent = 1;
    }
  vrp_start_timer(vrp_out_cell_translate(frame, seq), outgoing);
}

static void 
vrp_out_frame_send_hdr(vrp_outgoing_t outgoing, 
			    vrp_out_frame_t frame)
{
  vrp_show(50, "frame = %i size = %i\n", frame->frame_num, frame->hdr.size);
  assert(frame->hdr.size < VRP_MAX_PACKET_SIZE);
  vrp_libc_sendto(outgoing->fd,
		  &frame->hdr.content,
		  VRP_MAX_PACKET_SIZE, 
		  0,
		  (struct sockaddr*)&outgoing->outgoing_addr, 
		  sizeof(outgoing->outgoing_addr));
  /*frame->hdr.cell.lost=0; false loss statistics */
  vrp_start_timer(&frame->hdr.cell, outgoing);
}


static void 
vrp_out_frame_hdr_construct(vrp_outgoing_t outgoing, 
			    vrp_out_frame_t frame)
{
  struct vrp_buffer_s*buffer;
  struct vrp_critical_cell_s*cs;
  int b, e;

  vrp_show(60, "frame = %i\n", frame->frame_num);
  buffer = frame->src;
  vrp_hdr_pack_std(frame->hdr.content, &frame->hdr.size);
  /* 'create frame' tag */
  vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		   VRP_TAG_FRAME_CREATE, frame->frame_num, buffer->size, frame->data_per_packet);
  
  /* loss tolerance parameters */
  if((buffer->max_cons_loss > 0) && (buffer->max_cons_loss < 32768))
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		       VRP_TAG_CONS_LOSS, buffer->max_cons_loss / frame->data_per_packet );
    }
  else if((outgoing->max_cons_loss > 0) && (outgoing->max_cons_loss < 32768))
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		       VRP_TAG_CONS_LOSS, outgoing->max_cons_loss / frame->data_per_packet);
    }
  if((buffer->max_loss_rate >=0) && (buffer->max_loss_rate < 256))
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		       VRP_TAG_LOSS_RATE, buffer->max_loss_rate);
    }
  else if((outgoing->max_loss_rate >=0) && (outgoing->max_loss_rate < 256))
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		       VRP_TAG_LOSS_RATE, outgoing->max_loss_rate);
    }
  if(outgoing->min_ref_rate >=1)
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size,
		       VRP_TAG_MIN_REFRESH_RATE, outgoing->min_ref_rate);
    }

  /* calculate critical packets */
  cs = buffer->critical;
  while(cs != NULL)
    {
      b = cs->begin / frame->data_per_packet;
      e = (cs->end - 1) / frame->data_per_packet;
      if((b >= 0) && (e <= frame->max_seq) && (b <= e))
	{
	  if(b == e)
	    {
	      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
			       VRP_TAG_CRITICAL_PACKET, b);
	    }
	  else
	    {
	      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
			       VRP_TAG_CRITICAL_INTERVAL, b, e);
	    }
	}
      cs = cs->next;
    }
  /* check 'end of frame' */
  if(outgoing->current_frame != NULL && outgoing->current_frame->state == VRP_OUT_FRAME_CLOSING)
    {
      vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		       VRP_TAG_FRAME_CLOSE, outgoing->current_frame->frame_num);
      vrp_out_frame_destroy(outgoing, outgoing->current_frame);
    }

  /* end of header */
  vrp_hdr_pack_tag(frame->hdr.content, &frame->hdr.size, 
		   VRP_TAG_LAST);
}

void vrp_outgoing_read(vrp_outgoing_t outgoing)
{
  vrp_outgoing_read_callback(outgoing->fd, outgoing);
}

int
vrp_outgoing_read_callback(int fd, void *arg)
{
  vrp_outgoing_t outgoing;
  int save_errno;
  int done;
  struct sockaddr from_addr;
  int from_len;

  vrp_show(50, "fd = %i\n", fd);
  outgoing = (vrp_outgoing_t )arg;

  vrp_enter();
  done = 0;
  while(!done)
    {
      from_len = sizeof(from_addr);
#ifdef VRP_STANDALONE
      vrp_exit();
#endif
      outgoing->n_read = vrp_libc_recvfrom(outgoing->fd,
					   (char*) outgoing->recv_buf,
					   VRP_MAX_PACKET_SIZE,
					   0,
					   &from_addr,
					   &from_len);
      save_errno = errno;
#ifdef VRP_STANDALONE
      vrp_enter();
#endif
      vrp_show(60, "%i bytes read\n", outgoing->n_read);
      if(outgoing->n_read > 0)
	{
	  vrp_outgoing_process_fragment(outgoing);
#ifdef VRP_MADELEINE
	  /* right trick  ??? */
	  done=1;
#endif
	}
      else if (save_errno == EINTR)
	{
	  /* try again */
	}
      else if ((save_errno == EAGAIN) || (save_errno == EWOULDBLOCK))
	{
	  done = 1;
	}
      else
	{
	  vrp_warning("recvfrom() errno=%d n_read=%d (%s)\n", 
		      save_errno, outgoing->n_read, strerror(save_errno));
	}
    }

  vrp_exit();
  return 1;
}


static void
vrp_outgoing_process_fragment(vrp_outgoing_t outgoing)
{
  vrp_hdr_t hdr;
  vrp_out_frame_t frame;
  int i;
  int seq;
  long int moy;
  char*last_byte, *c, *prev;
  struct timeval timestamp;
  int sec, usec;
  struct vrp_out_cell_s*cell;

  gettimeofday(&timestamp, NULL);

  vrp_hdr_unpack(outgoing->recv_buf, &hdr);
  vrp_show(40, "received frame=%i seq=%i\n", hdr.frame, hdr.seq);
  c = (char*)outgoing->recv_buf + VRP_HDR_SIZE;
  last_byte = (char*)outgoing->recv_buf + outgoing->n_read - 1;
  /*
  if(hdr.frame != VRP_NO_FRAME)
    frame = vrp_out_frame_lookup(outgoing, hdr.frame);
  else
    frame = NULL;
  */
  if(outgoing->current_frame != NULL && outgoing->current_frame->frame_num == hdr.frame)
    frame = outgoing->current_frame;
  else
    {
      frame = NULL;
      vrp_show(50, "received fragment for no frame (hdr frame=%d current_frame=%p)\n", 
		 hdr.frame, outgoing->current_frame);
    }

  if(!hdr.is_hdr)
    goto badheader;

  while(*c != VRP_TAG_LAST)
    {
      if(c > last_byte)
	goto badheader;
      
      prev = c;
      
      if(frame != NULL)
	{
	  switch(*c)
	    {
	    case VRP_TAG_ACK:
	      if(c + 4 > last_byte)
		goto badheader;

	      vrp_mutex_lock(outgoing->mutex);
	      if (outgoing->current_frame!=NULL)		
		{
		  seq = vrp_unpack_int32(c+1);
	      
		  vrp_show(50, "ACK frame=%i packet=%i\n", frame->frame_num, seq);
		  if((frame->ack_expected <= seq) && 
		     (seq < frame->next_to_send))
		    {
		      cell = vrp_out_cell_translate(frame, seq);
		      cell->ack_time = timestamp;
		      for(i = frame->ack_expected; i <= seq; i++)
			{
			  vrp_out_cell_translate(frame, i)->acked = 1;
			}
		      frame->ack_expected = seq + 1;
		      vrp_out_frame_alive(frame);
		      if(!cell->expired)
			{
			  vrp_outgoing_adapt_timeout(outgoing, cell);
			}
		    }
		  else if(seq >= frame->next_to_send)
		    {
		      frame->state = VRP_OUT_FRAME_ERROR;
		    }
		  if (outgoing->will_slow_start)
		    {
		      vrp_show(30,"ACK received : I'm going to slowstart :%d\n",outgoing->will_slow_start_size);
		      outgoing->ack_cg_num++;
		      if (outgoing->ack_cg_num >= outgoing->cwnd_size/5 )
			{
			  vrp_mutex_unlock(outgoing->mutex);
			  vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size-1);
			  outgoing->ack_cg_num=0;
			}
		      else
			{
			  vrp_mutex_unlock(outgoing->mutex);
			}		      
		      if ( outgoing->cwnd_size <= outgoing->will_slow_start_size )
			outgoing->will_slow_start=0;
		    }
		  else if (outgoing->cwnd_size < outgoing->ssthresh)
		    {
		      vrp_show(30,"ACK received during slowstart < %d\n",outgoing->ssthresh);
		      vrp_mutex_unlock(outgoing->mutex);
		      vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size+1);
		    }
		  else
		    {		      
		      int i;
		      long int tmp;
		      outgoing->ack_cg_num++;
		      vrp_show(30,"ACK received during congestion avoidance\n");
#ifdef VEGAS_ENABLED
		      vrp_log(vrp_logv_fd, "%ld %d %d\n",outgoing->current_frame->next_to_send,outgoing->current_frame->frame_num,outgoing->avg_rtt_diff);
		      vrp_mutex_unlock(outgoing->mutex);
		      if (outgoing->ack_cg_num >= outgoing->cwnd_size )
		      {

		      if(outgoing->avg_rtt_diff>VRP_OUTGOING_VEGAS_DECREASE_THRESHOLD/outgoing->base_rtt)
			vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size-1);
		      else if (outgoing->avg_rtt_diff<VRP_OUTGOING_VEGAS_INCREASE_THRESHOLD/outgoing->base_rtt)
			vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size+1);
		      else
			vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size);

		      outgoing->ack_cg_num=0;
		      }
#else /* common congestion avoidance */

		      if (outgoing->ack_cg_num >= outgoing->cwnd_size )
			{
			  vrp_mutex_unlock(outgoing->mutex);
			  vrp_outgoing_adapt_cwd_size(outgoing,outgoing->cwnd_size+1);
			  outgoing->ack_cg_num=0;
			}
		      else
			{
			  vrp_mutex_unlock(outgoing->mutex);
			}		      
#endif
		    }
		}
	      else
		{
		  vrp_mutex_unlock(outgoing->mutex);
		  goto nullframe;
		}


	      c += 5;
	      break;
	    case VRP_TAG_NACK:
	      if(c + 4 > last_byte)
		goto badheader;
	      
	      vrp_mutex_lock(outgoing->mutex);
	      if (outgoing->current_frame!=NULL)
		{
		  seq = vrp_unpack_int32(c+1);

		  vrp_show(50, "NACK frame=%i packet=%i\n", frame->frame_num, seq);
		  if((frame->ack_expected <= seq) &&
		     (seq < frame->next_to_send))
		    {
		      vrp_outgoing_send_packet(outgoing, frame, hdr.seq);
		    }
		  else if(seq >= frame->next_to_send)
		    {
		      frame->state = VRP_OUT_FRAME_ERROR;
		    }
		  vrp_mutex_unlock(outgoing->mutex);
		  vrp_adapt_ssthresh(outgoing);
		}
	      else
		{
		  vrp_mutex_unlock(outgoing->mutex);
		  goto nullframe;
		}
	      c += 5;
	      break;
	    case VRP_TAG_SINGLE_ACK:
	      if(c + 4 > last_byte)
		goto badheader;

	      vrp_mutex_lock(outgoing->mutex);
	      if (outgoing->current_frame!=NULL)
		{
		  seq = vrp_unpack_int32(c+1);

		  vrp_show(50, "SACK frame=%i packet=%i\n", frame->frame_num, seq);
		  if((frame->ack_expected <= seq) && 
		     (seq < frame->next_to_send))
		    {
		      cell = vrp_out_cell_translate(frame, seq);
		      cell->acked = 1;
		      cell->ack_time = timestamp;
		      if(!cell->expired)
			{
			  vrp_outgoing_adapt_timeout(outgoing, cell);
			}
		      vrp_out_frame_alive(frame);
		    }
		  else if(seq >= frame->next_to_send)
		    {
		      frame->state = VRP_OUT_FRAME_ERROR;
		    }
		  vrp_mutex_unlock(outgoing->mutex);
		}
	      else
		{
		  vrp_mutex_unlock(outgoing->mutex);
		  goto nullframe;
		}	      

	      c += 5;
	      break;
	    default:
	      break;
	    }
	  /* update ACK info */
	  //vrp_mutex_lock(outgoing->mutex);
	  if (outgoing->current_frame==NULL)
	    {
	      goto nullframe;
	    }
	  else
	    {	  
	      while((frame->ack_expected < frame->next_to_send) && 
		    (vrp_out_cell_translate(frame, frame->ack_expected)->acked))
		{
		  frame->ack_expected++;
		  assert(frame->ack_expected <= frame->max_seq + 1);
		}
	      if(frame->ack_expected > frame->max_seq)
		{
		  frame->state = VRP_OUT_FRAME_CLOSING;
		  vrp_show(50, "CLOSE frame=%i\n", frame->frame_num);
		}
	    }
	  //vrp_mutex_unlock(outgoing->mutex);
	}
      switch(*c)
	{
	case VRP_TAG_PONG:
	  if(c + 8 > last_byte)
	    goto badheader;
	  
	  if(outgoing->state == VRP_OUTGOING_INIT)
	    {
	      sec = vrp_unpack_int32(c+1);
	      usec = vrp_unpack_int32(c+5);
	      outgoing->timing.RTT = (timestamp.tv_usec - usec + 1000000 * (timestamp.tv_sec - sec));
	      outgoing->timing.average_deviation = outgoing->timing.RTT / 4;
	      outgoing->timing.timeout = outgoing->timing.RTT + 4 * outgoing->timing.average_deviation;
	      if((outgoing->timing.timeout > 0) && 
		 (outgoing->timing.timeout < VRP_TIMEOUT_MAX))
		{
		  outgoing->state = VRP_OUTGOING_OK;
		  vrp_show(50, "timeout = %d usec\n", (int)outgoing->timing.timeout);
		}
	    }
	  c += 9;
	  break;
	case VRP_TAG_FRAME_SET:
	  {
	    int frame_num;
	    if(c + 2 > last_byte)
		goto badheader;
	    
	    frame_num = vrp_unpack_int16(c+1);
	    /*	  frame = vrp_out_frame_lookup(outgoing, seq); */
	    if(outgoing->current_frame != NULL && outgoing->current_frame->frame_num == frame_num)
	      {
		frame = outgoing->current_frame;
	      }
	    else
	      {
		frame = NULL;
		goto badheader;
	      }
	    c += 3;
	  }
  	  break;
	case VRP_TAG_FRAME_OK:
	  {
	    int frame_num;
	    if(c + 2 > last_byte)
		goto badheader;
	    
	    frame_num = vrp_unpack_int16(c+1);
	    /*	  frame = vrp_out_frame_lookup(outgoing, seq); */
	    if(outgoing->current_frame != NULL && 
	       outgoing->current_frame->frame_num == frame_num)
	      {
		frame = outgoing->current_frame;
		vrp_show(50, "FRAME_OK frame=%i accepted\n", frame_num);
		frame->state = VRP_OUT_FRAME_SEND;
	      }
	    else
	      {
		vrp_warning("received FRAME_OK for other frame than current-- current=%d received=%d\n",
			       outgoing->current_frame->frame_num, hdr.frame);
		frame = NULL;
		goto badheader;
	      }
	    c += 3;
	  }
	  break;
	case VRP_TAG_FRAME_REFUSED:
	  if(c + 2 > last_byte)
	      goto badheader;
	  
	  seq = vrp_unpack_int16(c+1);
	  vrp_show(20, "FRAME_REFUSED frame=%i refused\n", seq);
	  /*	  frame = vrp_out_frame_lookup(outgoing, seq); */
	  if(outgoing->current_frame != NULL && outgoing->current_frame->frame_num == hdr.frame)
	    frame = outgoing->current_frame;
	  else
	    frame = NULL;
	  if(frame != NULL)
	    frame->state = VRP_OUT_FRAME_REFUSED;
	  else
	    goto badheader;
	  
	  c += 3;
	  break;
	case VRP_TAG_FRAME_MISSING_HDR:
	  if(c + 2 > last_byte)
	      goto badheader;
	  
	  seq = vrp_unpack_int16(c+1);
	  /*	  frame = vrp_out_frame_lookup(outgoing, seq); */
	  if(outgoing->current_frame != NULL && outgoing->current_frame->frame_num == hdr.frame)
	    frame = outgoing->current_frame;
	  else
	    frame = NULL;
	  vrp_show(20, "MISSING_HDR frame=%i unknown\n", seq);
	  if(frame != NULL)
	    vrp_out_frame_send_hdr(outgoing, frame);
	  else
	    goto badheader;
	  
	  c += 3;
	  break;
	default:
	  break;
	}
      if(prev == c)
	goto badheader;
    }
  return;
badheader:
  vrp_show(20, "Outgoing -- received bad header (frame=%i)\n", hdr.frame);
  if (outgoing->current_frame==NULL)
    vrp_show(20, "Outgoing -- AND current_frame = NULL \n");
  return;
nullframe:
  vrp_show(20, "Outgoing -- current_frame = NULL \n");
  return;
}


static void 
vrp_outgoing_send_routine(vrp_outgoing_t outgoing, 
			  vrp_out_frame_t frame)
{
  struct vrp_out_cell_s*cell;
  int i;
  int first_timeout;

  vrp_show(60, "ack_expected=%i; next_to_send=%i; next_to_buffer=%i; max_seq=%i state=%i\n", 
	     frame->ack_expected, frame->next_to_send, 
	     frame->next_to_buffer, frame->max_seq, frame->state);

  vrp_mutex_lock(outgoing->mutex);
  /* clean buffers */
  while((frame->next_to_buffer < frame->ack_expected + frame->num_cells - 1) &&
	(frame->next_to_buffer <= frame->max_seq))
    {
      cell = vrp_out_cell_translate(frame, frame->next_to_buffer);
      cell->acked = 0;
      cell->sent = 0;
      cell->expired = 0;
      frame->next_to_buffer++;
    }
  vrp_mutex_unlock(outgoing->mutex);

  vrp_mutex_lock(outgoing->mutex);
  /* check ACK timeout */
  first_timeout = 1;
  for(i = frame->ack_expected; i < frame->next_to_send; i++)
    {
      cell = vrp_out_cell_translate(frame, i);
      if(!cell->acked)
	if(vrp_check_timeout(cell))
	  {
	    vrp_show(30, "Timeout expired packet=%i/%i frame=%i (timeout=%i usec)\n", 
		       i, frame->max_seq, frame->frame_num, (int)outgoing->timing.timeout);
	    vrp_show(30, "next_to_send=%d next_to_buffer=%d ack_expected=%d \n",
		       frame->next_to_send, frame->next_to_buffer, frame->ack_expected);
	    
	    vrp_mutex_unlock(outgoing->mutex);
	    if (!outgoing->will_slow_start && outgoing->cwnd_size > outgoing->ssthresh)
	      vrp_outgoing_slowstart(outgoing);
	    vrp_mutex_lock(outgoing->mutex);

	    if(first_timeout)
	      {
		vrp_outgoing_timeout_expired(outgoing);
		first_timeout = 0;
	      }
	    vrp_show(40, "re-send packet=%d frame=%d next_to_send=%d ack_expected=%d\n",
		       i, frame->frame_num, frame->next_to_send, frame->ack_expected);
	    vrp_outgoing_send_packet(outgoing, frame, i);
	  }
    }
  vrp_mutex_unlock(outgoing->mutex);

  vrp_mutex_lock(outgoing->mutex);
  /* send packets */
  while((frame->next_to_send < frame->next_to_buffer) &&
	(frame->next_to_send < frame->ack_expected + outgoing->fwnd_size) &&
	(frame->next_to_send < frame->ack_expected + outgoing->cwnd_size))
    {
      vrp_show(50, "Send packet=%d frame=%d\n", frame->next_to_send, frame->frame_num);
      vrp_outgoing_send_packet(outgoing, frame, frame->next_to_send);
      frame->next_to_send++;
    }
  vrp_mutex_unlock(outgoing->mutex);
}



static void vrp_start_timer(struct vrp_out_cell_s*cell, vrp_outgoing_t outgoing)
{
  gettimeofday(&cell->start_date, NULL);

  /* compute 'expire_date' */
  cell->expire_date.tv_sec  = cell->start_date.tv_sec;
  cell->expire_date.tv_usec = cell->start_date.tv_usec + outgoing->timing.timeout;
  while(cell->expire_date.tv_usec >= 1000000)
    {
      cell->expire_date.tv_sec  += 1;
      cell->expire_date.tv_usec -= 1000000;
    }
 
  /* compute 'expected' */
  cell->expected.tv_sec  = cell->start_date.tv_sec;
  cell->expected.tv_usec = cell->start_date.tv_usec + outgoing->timing.RTT;
  while(cell->expected.tv_usec >= 1000000)
    {
      cell->expected.tv_sec  += 1;
      cell->expected.tv_usec -= 1000000;
    }
  
  cell->timing.timeout = outgoing->timing.timeout;
  cell->timing.RTT = outgoing->timing.RTT;
  cell->timing.average_deviation = outgoing->timing.average_deviation;  
}

static int vrp_check_timeout(struct vrp_out_cell_s*cell)
{
  struct timeval date;
  
  gettimeofday(&date, NULL);
  return(((cell->expire_date.tv_sec == date.tv_sec) &&
	  (cell->expire_date.tv_usec < date.tv_usec)) ||
	 (cell->expire_date.tv_sec < date.tv_sec));
}

