/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

/*
 * Mad_mx.c
 * =========
 */


#include "madeleine.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <myriexpress.h>


/*
 * local structures
 * ----------------
 */
typedef struct s_mad_mx_driver_specific
{
        int dummy;
} mad_mx_driver_specific_t, *p_mad_mx_driver_specific_t;

typedef struct s_mad_mx_adapter_specific
{
        uint64_t	id;
} mad_mx_adapter_specific_t, *p_mad_mx_adapter_specific_t;

typedef struct s_mad_mx_channel_specific
{
        mx_endpoint_addr_t	endpoint_addr;
        mx_endpoint_t		endpoint;
        uint32_t		endpoint_id;
} mad_mx_channel_specific_t, *p_mad_mx_channel_specific_t;

typedef struct s_mad_mx_connection_specific
{
        uint32_t		remote_endpoint_id;
        mx_endpoint_addr_t	remote_endpoint_addr;
} mad_mx_connection_specific_t, *p_mad_mx_connection_specific_t;

typedef struct s_mad_mx_link_specific
{
        int dummy;
} mad_mx_link_specific_t, *p_mad_mx_link_specific_t;


/*
 * static functions
 * ----------------
 */

static
void
mad_mx_print_status(mx_status_t s) {
        DISP("status");

        switch(s.code) {
        case MX_STATUS_SUCCESS:
                DISP("Successful completion");
                break;
        case MX_STATUS_PENDING:
                DISP("Request still pending");
                break;
        case MX_STATUS_BUFFERED:
                DISP("Request has been buffered, but still pending");
                break;
        case MX_STATUS_REJECTED:
                DISP("Posted operation failed");
                break;
        case MX_STATUS_TIMEOUT:
                DISP("Posted operation timed out");
                break;
        case MX_STATUS_TRUNCATED:
                DISP("Operation completed, but data was truncated due to undersized buffer");
                break;
        case MX_STATUS_CANCELLED:
                DISP("Panding receive was cancelled");
                break;
        case MX_STATUS_ENDPOINT_UNKNOWN:
                DISP("Destination endpoint is unknown on the network fabric");
                break;
        case MX_STATUS_ENDPOINT_CLOSED:
                DISP("remoted endpoint is closed");
                break;
        case MX_STATUS_ENDPOINT_UNREACHABLE:
                DISP("Connectivity is broken between the source and the destination");
                //case MX_STATUS_BAD_SESSION:
                //DISP("Bad session (no mx_connect done?)"
                break;
        case MX_STATUS_BAD_KEY:
                DISP("Connect failed because of bad credentials");
                break;
        }
}

static
void
mad_mx_check_status(char *msg, mx_return_t return_code) {
        if (return_code != MX_SUCCESS) {
                DISP("%s failed with code %d/0x%x", msg, return_code, return_code);
                FAILURE("mx error");
        }
}

static
void
mad_mx_startup_info(void) {
        mx_return_t	 return_code	= MX_SUCCESS;
        uint32_t 	 nb_nic		= 0;
        uint64_t	*nic_id_array	= NULL;
        uint32_t	 nb_native_ep	= 0;
        uint32_t 	 i		= 0;

        LOG_IN();
        DISP("Madeleine/MX: hardware configuration");

        return_code = mx_get_info(NULL, MX_NIC_COUNT, &nb_nic, sizeof(nb_nic));
        mad_mx_check_status("mad_mx_startup_info", return_code);
        DISP("NIC count: %d", nb_nic);

        nic_id_array = TBX_MALLOC((nb_nic+1) * sizeof(*nic_id_array));
        mad_mx_check_status("mad_mx_startup_info", return_code);

        return_code = mx_get_info(NULL, MX_NIC_IDS, nic_id_array, (nb_nic+1) * sizeof(*nic_id_array));
        mad_mx_check_status("mad_mx_startup_info", return_code);

        for (i = 0; i < nb_nic; i++) {
                DISP("NIC %d id: %llx", i, nic_id_array[i]);
        }

        return_code = mx_get_info(NULL, MX_MAX_NATIVE_ENDPOINTS, &nb_native_ep, sizeof(nb_native_ep));
        mad_mx_check_status("mad_mx_startup_info", return_code);
        DISP("Native endpoint count: %d", nb_native_ep);

        LOG_OUT();
}



/*
 * Registration function
 * ---------------------
 */


void
mad_mx_register(p_mad_driver_t driver) {
        p_mad_driver_interface_t interface = NULL;

        LOG_IN();
        TRACE("Registering MX driver");
        interface = driver->interface;

        driver->connection_type  = mad_bidirectional_connection;
        driver->buffer_alignment = 32;
        driver->name             = tbx_strdup("mx");

        interface->driver_init                = mad_mx_driver_init;
        interface->adapter_init               = mad_mx_adapter_init;
        interface->channel_init               = mad_mx_channel_init;
        interface->before_open_channel        = NULL;
        interface->connection_init            = mad_mx_connection_init;
        interface->link_init                  = mad_mx_link_init;
        interface->accept                     = mad_mx_accept;
        interface->connect                    = mad_mx_connect;
        interface->after_open_channel         = NULL;
        interface->before_close_channel       = NULL;
        interface->disconnect                 = mad_mx_disconnect;
        interface->after_close_channel        = NULL;
        interface->link_exit                  = NULL;
        interface->connection_exit            = NULL;
        interface->channel_exit               = NULL;
        interface->adapter_exit               = NULL;
        interface->driver_exit                = NULL;
        interface->choice                     = NULL;
        interface->get_static_buffer          = NULL;
        interface->return_static_buffer       = NULL;
        interface->new_message                = mad_mx_new_message;
        interface->finalize_message           = NULL;
#ifdef MAD_MESSAGE_POLLING
        interface->poll_message               = mad_mx_poll_message;
#endif // MAD_MESSAGE_POLLING
        interface->receive_message            = mad_mx_receive_message;
        interface->message_received           = NULL;
        interface->send_buffer                = mad_mx_send_buffer;
        interface->receive_buffer             = mad_mx_receive_buffer;
        interface->send_buffer_group          = mad_mx_send_buffer_group;
        interface->receive_sub_buffer_group   = mad_mx_receive_sub_buffer_group;
        LOG_OUT();
}


void
mad_mx_driver_init(p_mad_driver_t d) {
        p_mad_mx_driver_specific_t	ds		= NULL;
        mx_return_t			return_code	= MX_SUCCESS;

        LOG_IN();
        TRACE("Initializing MX driver");
        ds  		= TBX_MALLOC(sizeof(mad_mx_driver_specific_t));
        d->specific	= ds;

        /* needed? */
        mx_set_error_handler(MX_ERRORS_RETURN);

        return_code	= mx_init();
        mad_mx_check_status("mx_init", return_code);

        /* debug only */
        mad_mx_startup_info();

        LOG_OUT();
}

void
mad_mx_adapter_init(p_mad_adapter_t a) {
        p_mad_mx_adapter_specific_t as = NULL;

        LOG_IN();
        as		= TBX_MALLOC(sizeof(mad_mx_adapter_specific_t));

        if (strcmp(a->dir_adapter->name, "default")) {
                FAILURE("unsupported adapter");
        } else {
                as->id	= MX_ANY_NIC;
        }

        a->specific	= as;
        a->parameter	= tbx_strdup("-");
        LOG_OUT();
}

void
mad_mx_channel_init(p_mad_channel_t ch) {
        p_mad_mx_adapter_specific_t 	as = NULL;
        p_mad_mx_channel_specific_t	chs		= NULL;
        mx_return_t			return_code	= MX_SUCCESS;
        uint32_t                        e_id     	= 1;
        const uint32_t			e_key		= 0xFFFFFFFF;
        mx_endpoint_t			e;
        mx_endpoint_addr_t		e_addr;
        LOG_IN();
        as =	ch->adapter->specific;

        /*
          Current prototype:

          mx_return_t mx_open_endpoint(uint64_t nic_id,
			     uint32_t endpoint_id,
			     uint32_t endpoint_key,
			     mx_param_t *params_array,
			     uint32_t params_count,
			     mx_endpoint_t *endpoint);
        */

        return_code	= mx_open_endpoint(as->id,
                                           e_id,
                                           e_key,
                                           NULL,
                                           0,
                                           &e);
        mad_mx_check_status("mx_open_endpoint", return_code);

        return_code	= mx_get_endpoint_addr(e, &e_addr);
        mad_mx_check_status("mx_get_endpoint", return_code);


        chs			= TBX_MALLOC(sizeof(mad_mx_channel_specific_t));
        chs->endpoint		= e;
        chs->endpoint_id	= e_id;
        ch->specific		= chs;
        LOG_OUT();
}

void
mad_mx_connection_init(p_mad_connection_t in,
                       p_mad_connection_t out) {
        p_mad_mx_connection_specific_t	cs	= NULL;
        p_mad_channel_t			ch	= NULL;
        uint32_t			re_id	= 1;

        LOG_IN();
        ch = in->channel;

        cs			= TBX_MALLOC(sizeof(mad_mx_connection_specific_t));
        cs->remote_endpoint_id	= re_id;

        in->specific		= cs;
        in->nb_link		= 1;

        out->specific		= cs;
        out->nb_link		= 1;
        LOG_OUT();
}

void
mad_mx_link_init(p_mad_link_t lnk) {
        LOG_IN();
        /* lnk->link_mode   = mad_link_mode_buffer_group; */
        lnk->link_mode   = mad_link_mode_buffer;
        lnk->buffer_mode = mad_buffer_mode_dynamic;
        lnk->group_mode  = mad_group_mode_split;
        LOG_OUT();
}

void
mad_mx_accept(p_mad_connection_t   in,
              p_mad_adapter_info_t ai TBX_UNUSED) {
        p_mad_mx_connection_specific_t	 is		= NULL;
        p_mad_mx_channel_specific_t	 chs		= NULL;
        mx_return_t			 return_code	= MX_SUCCESS;
        uint32_t			 nb_r_nic_ids	= 1;
        uint64_t			 r_nic_id	= 0;
        const uint32_t			 e_key		= 0xFFFFFFFF;
        char				*r_hostname	= NULL;
        size_t				 r_hostname_len	= 0;
        mx_endpoint_addr_t 		 re_addr;

        LOG_IN();
        is	= in->specific;
        chs	= in->channel->specific;

        r_hostname_len	= strlen(ai->dir_node->name) + 2;
        r_hostname	= TBX_MALLOC(r_hostname_len + 1);
        strcpy(r_hostname, ai->dir_node->name);

        {
                char *tmp = NULL;

                tmp = strchr(r_hostname, '.');
                if (tmp) {
                        *tmp = '\0';
                }
        }

        strcat(r_hostname, ":0");

        /*
          Current prototype:

mx_return_t
mx_hostname_to_nic_ids(char *hostname,
		       uint64_t *nic_ids,
		       uint32_t *num_nic_ids)
        */
        return_code	= mx_hostname_to_nic_ids(r_hostname, &r_nic_id, &nb_r_nic_ids);
        mad_mx_check_status("mx_hostname_to_nic_ids", return_code);

        TBX_FREE(r_hostname);
        r_hostname	= NULL;
        r_hostname_len	= 0;

        /*
          Current prototype:

mx_return_t
mx_connect(struct mx_endpoint *ep, uint64_t nic_id, uint32_t eid,
	   uint32_t key, uint32_t timeout, mx_endpoint_addr_t *addrp)
        */
        return_code	= mx_connect(chs->endpoint,
                                     r_nic_id,
                                     is->remote_endpoint_id,
                                     e_key,
                                     MX_INFINITE,
                                     &re_addr);

        mad_mx_check_status("mx_connect", return_code);
        is->remote_endpoint_addr	= re_addr;
        LOG_OUT();
}

void
mad_mx_connect(p_mad_connection_t   out,
               p_mad_adapter_info_t ai) {
        p_mad_mx_connection_specific_t	 os		= NULL;
        p_mad_mx_channel_specific_t	 chs		= NULL;
        mx_return_t			 return_code	= MX_SUCCESS;
        uint32_t			 nb_r_nic_ids	= 1;
        uint64_t			 r_nic_id	= 0;
        const uint32_t			 e_key		= 0xFFFFFFFF;
        char				*r_hostname	= NULL;
        size_t				 r_hostname_len	= 0;
        mx_endpoint_addr_t 		 re_addr;

        LOG_IN();
        os	= out->specific;
        chs	= out->channel->specific;

        r_hostname_len	= strlen(ai->dir_node->name) + 2;
        r_hostname	= TBX_MALLOC(r_hostname_len + 1);
        strcpy(r_hostname, ai->dir_node->name);

        {
                char *tmp = NULL;

                tmp = strchr(r_hostname, '.');
                if (tmp) {
                        *tmp = '\0';
                }
        }

        strcat(r_hostname, ":0");

        //DISP_STR("r_hostname", r_hostname);
        //fprintf(stderr, "r_hostname = %s\n", r_hostname);

        /*
          Current prototype:

mx_return_t
mx_hostname_to_nic_ids(char *hostname,
		       uint64_t *nic_ids,
		       uint32_t *num_nic_ids)
        */
        return_code	= mx_hostname_to_nic_ids(r_hostname, &r_nic_id, &nb_r_nic_ids);
        mad_mx_check_status("mx_hostname_to_nic_ids", return_code);

        TBX_FREE(r_hostname);
        r_hostname	= NULL;
        r_hostname_len	= 0;


        /*
          Current prototype:

mx_return_t
mx_connect(struct mx_endpoint *ep, uint64_t nic_id, uint32_t eid,
	   uint32_t key, uint32_t timeout, mx_endpoint_addr_t *addrp)
        */
        return_code	= mx_connect(chs->endpoint,
                                     r_nic_id,
                                     os->remote_endpoint_id,
                                     e_key,
                                     MX_INFINITE,
                                     &re_addr);
        mad_mx_check_status("mx_connect", return_code);
        os->remote_endpoint_addr	= re_addr;
        LOG_OUT();
}

void
mad_mx_disconnect(p_mad_connection_t c) {
        p_mad_mx_connection_specific_t cs = NULL;

        LOG_IN();
        cs = c->specific;
        LOG_OUT();
}

void
connection_exit(p_mad_connection_t in,
                p_mad_connection_t out) {
        p_mad_mx_connection_specific_t cs = NULL;

        LOG_IN();
        cs = in->specific;
        TBX_FREE(cs);
        in->specific	= NULL;
        out->specific	= NULL;
        LOG_OUT();
}

void
mad_mx_channel_exit(p_mad_channel_t ch) {
        p_mad_mx_channel_specific_t	chs		= NULL;
        mx_return_t			return_code	= MX_SUCCESS;

        LOG_IN();
        chs = ch->specific;
        return_code	= mx_close_endpoint(chs->endpoint);
        mad_mx_check_status("mx_close_endpoint", return_code);
        TBX_FREE(chs);
        ch->specific	= NULL;
        LOG_OUT();
}

void
mad_mx_adapter_exit(p_mad_adapter_t a) {
        p_mad_mx_adapter_specific_t as = NULL;

        LOG_IN();
        as		= a->specific;
        TBX_FREE(as);
        a->specific	= NULL;
        a->parameter	= NULL;
        LOG_OUT();
}

void
mad_mx_driver_exit(p_mad_driver_t d) {
        p_mad_mx_driver_specific_t	ds		= NULL;
        mx_return_t			return_code	= MX_SUCCESS;

        LOG_IN();
        TRACE("Finalizing MX driver");
        ds	= d->specific;
        TBX_FREE(ds);
        d->specific	= NULL;

        return_code	= mx_finalize();
        mad_mx_check_status("mx_finalize", return_code);
        LOG_OUT();
}


#ifdef MAD_MESSAGE_POLLING
p_mad_connection_t
mad_mx_poll_message(p_mad_channel_t channel)
{
        FAILURE("unsupported");
}
#endif // MAD_MESSAGE_POLLING

void
mad_mx_new_message(p_mad_connection_t out){

        p_mad_mx_connection_specific_t	os	= NULL;
        p_mad_mx_channel_specific_t	chs		= NULL;
        mx_endpoint_t ep = NULL;
        mx_endpoint_addr_t dest_address ;
        mx_request_t request = 0;
        mx_status_t status;
        uint32_t result = 0;
        mx_return_t rc = MX_SUCCESS;
        uint64_t match_info = 0;
        //        mx_segment_t seg;

        LOG_IN();
        os	= out->specific;
        chs	= out->channel->specific;

        ep = chs->endpoint;

        dest_address = os->remote_endpoint_addr;



        //seg.segment_ptr = buffer->buffer+buffer->bytes_read;
        //seg.segment_length = buffer->bytes_written - buffer->bytes_read;

        match_info = (uint64_t)out->channel->id <<32 | out->channel->process_lrank;

        /*
       mx_return_t
mx_isend(struct mx_endpoint * ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
#ifdef MX_KERNEL
	 uint32_t pin_type,
#endif
	 mx_endpoint_addr_t dest_address,
         uint64_t match_info,
	 void *context, mx_request_t *request)
        */

        rc = mx_isend(ep,
                      //&seg, 1,
                      NULL, 0,
                      dest_address,
                      match_info,
                      NULL, &request);

        mad_mx_check_status("mx_isend", rc);


        //rc = mx_wait(ep, &request, MX_INFINITE, &status, &result);
	do {
		rc = mx_test(ep, &request, &status, &result);
	} while (rc == MX_SUCCESS && !result);

        mad_mx_check_status("mx_wait", rc);
        //mad_mx_print_status(status);

        LOG_OUT();
}




p_mad_connection_t
mad_mx_receive_message(p_mad_channel_t ch) {
        p_mad_mx_channel_specific_t	chs		= NULL;
        p_tbx_darray_t			in_darray	= NULL;
        p_mad_connection_t		in		= NULL;
        mx_endpoint_t ep = NULL;
        mx_request_t request = 0;
        mx_status_t status;
        uint32_t result = 0;
        mx_return_t rc = MX_SUCCESS;
        uint64_t match_info = 0;
        //        mx_segment_t seg;

        LOG_IN();
        chs		= ch->specific;
        in_darray	= ch->in_connection_darray;




        ep = chs->endpoint;


        //seg.segment_ptr = buffer->buffer + buffer->bytes_written;
        //seg.segment_length = buffer->length - buffer->bytes_written;

        match_info = (uint64_t)ch->id <<32;


        /*
mx_return_t
mx_irecv(mx_endpoint_t ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
#ifdef MX_KERNEL
	 uint32_t pin_type,
#endif
	 uint64_t match_info, uint64_t match_mask,
	 void *context, mx_request_t *request)
        */

        rc = mx_irecv(ep,
                      NULL, 0,
                      //&seg, 1,
                 match_info,
                 MX_MATCH_MASK_BC, NULL, &request);
        mad_mx_check_status("mx_irecv", rc);


        //rc = mx_wait(ep, &request, MX_INFINITE, &status, &result);
	do {
		rc = mx_test(ep, &request, &status, &result);
	} while (rc == MX_SUCCESS && !result);
        mad_mx_check_status("mx_wait", rc);
        //mad_mx_print_status(status);

        in = tbx_darray_get(in_darray, status.match_info & 0xFFFFFFFF);

        LOG_OUT();

        return in;
}

void
mad_mx_send_buffer(p_mad_link_t     lnk,
                   p_mad_buffer_t   buffer) {
        p_mad_connection_t		out	= NULL;
        p_mad_mx_connection_specific_t	os	= NULL;
        p_mad_mx_channel_specific_t	chs		= NULL;
        mx_endpoint_t ep = NULL;
        mx_endpoint_addr_t dest_address ;
        mx_request_t request = 0;
        mx_status_t status;
        uint32_t result = 0;
        mx_return_t rc = MX_SUCCESS;
        uint64_t match_info = 0;
        mx_segment_t seg;

        LOG_IN();
        out	= lnk->connection;
        os	= out->specific;
        chs	= out->channel->specific;

        ep = chs->endpoint;

        dest_address = os->remote_endpoint_addr;

        seg.segment_ptr = buffer->buffer+buffer->bytes_read;
        seg.segment_length = buffer->bytes_written - buffer->bytes_read;

        match_info = (uint64_t)out->channel->id <<32 | out->channel->process_lrank;

        /*
       mx_return_t
mx_isend(struct mx_endpoint * ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
#ifdef MX_KERNEL
	 uint32_t pin_type,
#endif
	 mx_endpoint_addr_t dest_address,
         uint64_t match_info,
	 void *context, mx_request_t *request)
        */

        rc = mx_isend(ep,
                 &seg, 1,
                 dest_address,
                 match_info,
                 NULL, &request);
        mad_mx_check_status("mx_isend", rc);


        //rc = mx_wait(ep, &request, MX_INFINITE, &status, &result);
	do {
		rc = mx_test(ep, &request, &status, &result);
	} while (rc == MX_SUCCESS && !result);
        mad_mx_check_status("mx_wait", rc);
        //        mad_mx_print_status(status);

        LOG_OUT();
}

void
mad_mx_receive_buffer(p_mad_link_t    lnk,
                      p_mad_buffer_t *_buffer) {
        p_mad_buffer_t			buffer  = *_buffer;
        p_mad_connection_t		in	= NULL;
        p_mad_mx_connection_specific_t	is	= NULL;
        p_mad_mx_channel_specific_t	chs		= NULL;
        mx_endpoint_t ep = NULL;
        mx_request_t request = 0;
        mx_status_t status;
        uint32_t result = 0;
        mx_return_t rc = MX_SUCCESS;
        uint64_t match_info = 0;
        mx_segment_t seg;

        LOG_IN();
        in	= lnk->connection;
        is	= in->specific;

        chs	= in->channel->specific;

        ep = chs->endpoint;


        seg.segment_ptr = buffer->buffer + buffer->bytes_written;
        seg.segment_length = buffer->length - buffer->bytes_written;

        match_info = (uint64_t)in->channel->id <<32 | in->remote_rank;


        /*
mx_return_t
mx_irecv(mx_endpoint_t ep,
	 mx_segment_t *segments_list, uint32_t segments_count,
#ifdef MX_KERNEL
	 uint32_t pin_type,
#endif
	 uint64_t match_info, uint64_t match_mask,
	 void *context, mx_request_t *request)
        */

        rc = mx_irecv(ep,
                 &seg, 1,
                 match_info,
                 MX_MATCH_MASK_NONE, NULL, &request);
        mad_mx_check_status("mx_irecv", rc);


        //rc = mx_wait(ep, &request, MX_INFINITE, &status, &result);
	do {
		rc = mx_test(ep, &request, &status, &result);
	} while (rc == MX_SUCCESS && !result);
        mad_mx_check_status("mx_wait", rc);
        //mad_mx_print_status(status);
        LOG_OUT();
}

void
mad_mx_send_buffer_group_1(p_mad_link_t         lnk,
                           p_mad_buffer_group_t buffer_group) {
        LOG_IN();
        if (!tbx_empty_list(&(buffer_group->buffer_list))) {
                tbx_list_reference_t ref;

                tbx_list_reference_init(&ref, &(buffer_group->buffer_list));
                do {
                        p_mad_buffer_t b = NULL;

                        b = tbx_get_list_reference_object(&ref);
                        mad_mx_send_buffer(lnk, b);
                } while(tbx_forward_list_reference(&ref));
        }
        LOG_OUT();
}

void
mad_mx_receive_sub_buffer_group_1(p_mad_link_t         lnk,
                                  p_mad_buffer_group_t buffer_group) {
        LOG_IN();
        if (!tbx_empty_list(&(buffer_group->buffer_list))) {
                tbx_list_reference_t ref;

                tbx_list_reference_init(&ref, &(buffer_group->buffer_list));
                do {
                        p_mad_buffer_t b = NULL;

                        b = tbx_get_list_reference_object(&ref);
                        mad_mx_receive_buffer(lnk, &b);
                } while(tbx_forward_list_reference(&ref));
        }
        LOG_OUT();
}

void
mad_mx_send_buffer_group(p_mad_link_t         lnk,
                         p_mad_buffer_group_t buffer_group) {
        LOG_IN();
        mad_mx_send_buffer_group_1(lnk, buffer_group);
        LOG_OUT();
}

void
mad_mx_receive_sub_buffer_group(p_mad_link_t         lnk,
                                tbx_bool_t           first_sub_group
                                __attribute__ ((unused)),
                                p_mad_buffer_group_t buffer_group) {
        LOG_IN();
        mad_mx_receive_sub_buffer_group_1(lnk, buffer_group);
        LOG_OUT();
}

