
/*
 * PM2: Parallel Multithreaded Machine
 * Copyright (C) 2001 "the PM2 team" (see AUTHORS file)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include "marcel.h"

static marcel_lock_t __polling_lock = MARCEL_LOCK_INIT;

static poll_struct_t poll_structs[MAX_POLL_IDS];
static unsigned nb_poll_structs = 0;

poll_struct_t *__polling_tasks = NULL;

// Checks to see if some polling jobs should be done. NOTE: The
// function assumes that:
//   1) "lock_task()" was called previously ;
//   2) __polling_lock is acquired
//   3) la structure poll_struct_t possde au moins une tche en attente de poll
int __marcel_check_polling_for(poll_struct_t *ps)
{
	int waked_some_task = 0;
	register poll_cell_t *cell;
	poll_struct_t *p;

        DEFINE_CUR_LWP(, __attribute__((unused)) =, GET_LWP(marcel_self()));

	if(ps->nb_cells == 1 && ps->fastfunc) {
		ps->cur_cell = ps->first_cell;
		cell = ((poll_cell_t *)
			(*ps->fastfunc)(ps, ps->first_cell->arg, FALSE) ?
			ps->first_cell : MARCEL_POLL_FAILED);
	} else {
		cell = (poll_cell_t *)(*ps->func)(ps,
						  SCHED_DATA(cur_lwp).running_tasks,
						  marcel_sleepingthreads(),
						  marcel_blockedthreads());
	}

	if(cell != MARCEL_POLL_FAILED) {

		waked_some_task = 1;
		marcel_wake_task(cell->task, &cell->blocked);
		mdebug("Poll succeed with task %p\n", cell->task);

		/* Retrait de d'un lment de la liste */
		if(cell->prev != NULL)
			cell->prev->next = cell->next;
		else
			ps->first_cell = cell->next;
		if(cell->next != NULL)
			cell->next->prev = cell->prev;
		ps->nb_cells--;
		
		if(!ps->nb_cells) {
			/* Il faut retirer la tache de __polling_task */
			if((p = ps->prev) == ps) {
				__polling_tasks = NULL;
			} else {
				if(ps == __polling_tasks)
					__polling_tasks = p;
				p->next = ps->next;
				p->next->prev = p;
			}
		} else {
			/* S'il reste au moins 2 requetes ou s'il n'y a pas de
			   "fast poll", alors il faut factoriser. */
			if(ps->nb_cells > 1 || !ps->fastfunc) {
				mdebug("Factorizing polling");
				(*(ps->gfunc))((marcel_pollid_t)ps);
			}
		}
	}
	return waked_some_task;
}

// Checks to see if some polling jobs should be done. NOTE: The
// function assumes that:
//   1) "lock_task()" was called previously ;
int __marcel_check_polling(unsigned polling_point)
{
	int waked_some_task = 0;
	poll_struct_t *ps;

	LOG_IN();

	if(marcel_lock_tryacquire(&__polling_lock)) {
		
		if((ps = __polling_tasks) == NULL) {
			marcel_lock_release(&__polling_lock);
			
			LOG_OUT();
			return 0;
		}

		do {
			if(ps->polling_points & polling_point) {
				waked_some_task = __marcel_check_polling_for(ps);
				if (__polling_tasks == NULL) {
					/* La dernire tche de polling vient de disparatre ... */
					break;
				}
			}
			ps = ps->next;
		} while(ps != __polling_tasks);
		
		marcel_lock_release(&__polling_lock);
	}
	
	LOG_OUT();

	return waked_some_task;
}

int marcel_force_check_polling(poll_struct_t *ps)
{
	int waked_some_task = 0;
	
	LOG_IN();
	
	lock_task();
	marcel_lock_acquire(&__polling_lock);
	
	if(ps->first_cell == NULL) {
		/* Rien en attente sur ce poll_struct_t */
		marcel_lock_release(&__polling_lock);
		LOG_OUT();
		return 0;
	}
		
	waked_some_task = __marcel_check_polling_for(ps);
	marcel_lock_release(&__polling_lock);
	unlock_task();

	LOG_OUT();

	return waked_some_task;
}

marcel_pollid_t marcel_pollid_create(marcel_pollgroup_func_t g,
				     marcel_poll_func_t f,
				     marcel_fastpoll_func_t h,
				     unsigned polling_points)
{
  marcel_pollid_t id;

  LOG_IN();

  lock_task();

  if(nb_poll_structs == MAX_POLL_IDS) {
    unlock_task();
    RAISE(CONSTRAINT_ERROR);
  }

  id = &poll_structs[nb_poll_structs++];

  unlock_task();

  id->first_cell = id->cur_cell = NULL;
  id->nb_cells = 0;
  id->gfunc = g;
  id->func = f;
  id->fastfunc = h;
  id->polling_points = polling_points | MARCEL_POLL_AT_IDLE;

  mdebug("registering pollid %p (gr=%p, func=%p, fast=%p, pts=%x)\n",
	 id, g, f, h, id->polling_points);

  LOG_OUT();

  return id;
}

void marcel_poll(marcel_pollid_t id, any_t arg)
{
  poll_cell_t cell;

  LOG_IN();
  mdebug("Marcel_poll (thread %p)...\n", marcel_self());
  mdebug("using pollid %p (gr=%p, func=%p, fast=%p, pts=%x)\n",
	 id, id->gfunc, id->func, id->fastfunc, id->polling_points);

  lock_task();

  marcel_lock_acquire(&__polling_lock);

  if(id->fastfunc) {
    mdebug("Using Immediate FastPoll %p\n", id->fastfunc);
    id->cur_cell = &cell;
    if((*id->fastfunc)(id, arg, TRUE) != MARCEL_POLL_FAILED) {
      mdebug("Fast Poll completed ok!\n");
      marcel_lock_release(&__polling_lock);
      unlock_task();
      return;
    }
  }

  cell.task = marcel_self();
  cell.blocked = TRUE;
  cell.arg = arg;

  /* Insertion dans la liste des taches sur le meme id */
  cell.prev = NULL;
  if((cell.next = id->first_cell) != NULL)
    cell.next->prev = &cell;
  id->first_cell = &cell;
  id->nb_cells++;

  /* Si pas de "fast poll" ou si au moins 2 requetes de poll, il faut
     factoriser... */
  if(!id->fastfunc || id->nb_cells > 1) {
    mdebug("Factorizing polling");
    (*(id->gfunc))(id);
  }

  /* Enregistrement ventuel dans la __polling_tasks */
  if(cell.next == NULL) {
    /* Il n'y avait pas encore un polling de ce type */
    if(__polling_tasks == NULL) {
      __polling_tasks = id;
      id->next = id->prev = id;
    } else {
      id->next = __polling_tasks;
      id->prev = __polling_tasks->prev;
      id->prev->next = __polling_tasks->prev = id;
    }
  }

  marcel_lock_release(&__polling_lock);
  marcel_give_hand(&cell.blocked);
  LOG_OUT();
}

